import argparse
import os
import time

import txaio
from autobahn.twisted.util import sleep as dsleep
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from twisted.internet.defer import inlineCallbacks

from socs.snmp import SNMPTwister

# For logging
txaio.use_twisted()


def _extract_oid_field_and_value(get_result):
    """Extract field names and OID values from SNMP GET results.

    The ObjectType objects returned from pysnmp interactions contain the
    info we want to use for field names, specifically the OID and associated
    integer for uniquely identifying duplicate OIDs, as well as the value of the
    OID, which we want to save.

    Here we use the prettyPrint() method to get the OID name, requiring
    some string manipulation. We also just grab the hidden ._value
    directly, as this seemed the cleanest way to get an actual value of a
    normal type. Without doing this we get a pysnmp defined Integer32 or
    DisplayString, which were awkward to handle, particularly the
    DisplayString.

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.

    Returns
    -------
    field_name : str
        Field name for an OID, i.e. 'outletStatus_1'
    oid_value : int or str
        Associated value for the OID. Returns None if not an int or str
    oid_description : str
        String description of the OID value.
    """
    # OID from SNMP GET
    oid = get_result[0].prettyPrint()
    # Makes something like 'IBOOTPDU-MIB::outletStatus.1'
    # look like 'outletStatus_1'
    field_name = oid.split("::")[1].replace('.', '_')

    # Grab OID value, mostly these are integers
    oid_value = get_result[1]._value
    oid_description = get_result[1].prettyPrint()

    # Decode string values
    if isinstance(oid_value, bytes):
        oid_value = oid_value.decode("utf-8")

    # I don't expect any other types at the moment, but just in case.
    if not isinstance(oid_value, (int, bytes, str)):
        oid_value = None

    return field_name, oid_value, oid_description


def _build_message(get_result, names, time):
    """Build the message for publication on an OCS Feed.

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.
    names : list
        List of strings for outlet names
    time : float
        Timestamp for when the SNMP GET was issued.

    Returns
    -------
    message : dict
        OCS Feed formatted message for publishing
    """
    message = {
        'block_name': 'ibootbar',
        'timestamp': time,
        'data': {}
    }

    for item in get_result:
        field_name, oid_value, oid_description = _extract_oid_field_and_value(item)

        if oid_value is None:
            continue

        message['data'][field_name] = oid_value
        message['data'][field_name + "_name"] = names[int(field_name[-1])]
        message['data'][field_name + "_description"] = oid_description

    return message


def update_cache(get_result, names, outlet_locked, timestamp):
    """Update the OID Value Cache.

    The OID Value Cache is used to store each unique OID and will be passed to
    session.data

    The cache consists of a dictionary, with the unique OIDs as keys, and
    another dictionary as the value. Each of these nested dictionaries contains
    the OID values, name, and description (decoded string). An example for a
    single OID, with connection status and timestamp information::

        {"outletStatus_0": {"status": 1,
                            "name": Outlet-1,
                            "description": "on"},
         "ibootbar_connection": {"last_attempt": 1598543359.6326838,
                                 "connected": True},
         "timestamp": 1656085022.680916}

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.
    names : list
        List of strings for outlet names
    outlet_locked : list of bool
        List of bool for outlets (1-8) that are locked
    timestamp : float
        Timestamp for when the SNMP GET was issued.
    """
    oid_cache = {}
    # Return disconnected if SNMP response is empty
    if get_result is None:
        oid_cache['ibootbar_connection'] = {'last_attempt': time.time(),
                                            'connected': False}
        return oid_cache

    for item in get_result:
        field_name, oid_value, oid_description = _extract_oid_field_and_value(item)
        if oid_value is None:
            continue

        # Update OID Cache for session.data
        oid_cache[field_name] = {"status": oid_value}
        oid_cache[field_name]["name"] = names[int(field_name[-1])]
        oid_cache[field_name]["locked"] = outlet_locked[int(field_name[-1])]
        oid_cache[field_name]["description"] = oid_description
        oid_cache['ibootbar_connection'] = {'last_attempt': time.time(),
                                            'connected': True}
        oid_cache['timestamp'] = timestamp

    return oid_cache


class ibootbarAgent:
    """Monitor the ibootbar system via SNMP.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    address : str
        Address of the ibootbar.
    port : int
        SNMP port to issue GETs to, default to 161.
    version : int
        SNMP version for communication (1, 2, or 3), defaults to 2.
    lock_outlet : list of ints
        List of outlets to lock on agent startup. Outlets are numbered 1-8.

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    is_streaming : bool
        Tracks whether or not the agent is actively issuing SNMP GET commands
        to the ibootbar. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    """

    def __init__(self, agent, address, port=161, version=2, lock_outlet=None):
        self.agent = agent
        self.is_streaming = False
        self.log = self.agent.log
        self.lock = TimeoutLock()

        # Initialize with locked outlets if given in args
        self.outlet_locked = [False for i in range(8)]
        if lock_outlet is not None:
            for i in range(8):
                if i + 1 in lock_outlet:
                    self.outlet_locked[i] = True
                else:
                    self.outlet_locked[i] = False

        self.log.info(f'Using SNMP version {version}.')
        self.version = version
        self.address = address
        self.snmp = SNMPTwister(address, port)

        self.lastGet = 0
        self.sample_period = 60

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('ibootbar',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('test_mode', default=False, type=bool)
    @inlineCallbacks
    def acq(self, session, params=None):
        """acq()

        **Process** - Fetch values from the ibootbar via SNMP.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            {'outletStatus_0':
                {'status': 1,
                 'name': 'Outlet-1',
                 'description': 'on'},
             'outletStatus_1':
                {'status': 0,
                 'name': 'Outlet-2',
                 'description': 'off'},
             ...
             'ibootbar_connection':
                {'last_attempt': 1656085022.680916,
                 'connected': True},
             'timestamp': 1656085022.680916,
             'address': '10.10.10.50'}
        """
        # Set initial default outlet names
        # Note outlets are numbered 1-8 physically on ibootbars, OIDs are numbered 0-7
        names = ['Outlet-1', 'Outlet-2', 'Outlet-3', 'Outlet-4',
                 'Outlet-5', 'Outlet-6', 'Outlet-7', 'Outlet-8']

        session.set_status('running')
        self.is_streaming = True
        while self.is_streaming:
            yield dsleep(1)
            read_time = time.time()

            # Check if sample period has passed before getting status
            if (read_time - self.lastGet) < self.sample_period:
                continue

            get_list = []
            name_list = []
            names = []

            # Create the lists of OIDs to send get commands
            for i in range(8):
                get_list.append(('IBOOTPDU-MIB', 'outletStatus', i))
                name_list.append(('IBOOTPDU-MIB', 'outletName', i))

            # Issue SNMP GET command
            get_result = yield self.snmp.get(get_list, self.version)
            name_result = yield self.snmp.get(name_list, self.version)

            # If device gets disconnected, name_result is None
            if name_result is not None:
                for item in name_result:
                    names.append(item[1].prettyPrint())

            # Do not publish if ibootbar connection has dropped
            try:
                # Update session.data
                oid_cache = update_cache(get_result, names, self.outlet_locked, read_time)
                oid_cache['address'] = self.address
                session.data = oid_cache
                self.log.debug("{data}", data=session.data)

                if get_result is None:
                    raise ConnectionError('No SNMP response. Check your connection.')

                self.lastGet = time.time()
                # Publish to feed
                message = _build_message(get_result, names, read_time)
                self.log.debug("{msg}", msg=message)
                session.app.publish_to_feed('ibootbar', message)
            except ConnectionError as e:
                self.log.error(f'{e}')
                yield dsleep(1)
                self.log.info('Trying to reconnect.')

            if params['test_mode']:
                break

        return True, "Finished Recording"

    def _stop_acq(self, session, params=None):
        """_stop_acq()
        **Task** - Stop task associated with acq process.
        """
        if self.is_streaming:
            session.set_status('stopping')
            self.is_streaming = False
            return True, "Stopping Recording"
        else:
            return False, "Acq is not currently running"

    @ocs_agent.param('outlet', choices=[1, 2, 3, 4, 5, 6, 7, 8])
    @ocs_agent.param('state', choices=['on', 'off'])
    @inlineCallbacks
    def set_outlet(self, session, params=None):
        """set_outlet(outlet, state)

        **Task** - Set a particular outlet to on/off.

        Parameters
        ----------
        outlet : int
            Outlet number to set. Choices are 1-8 (physical outlets).
        state : str
            State to set outlet to, which may be 'on' or 'off'
        """
        with self.lock.acquire_timeout(3, job='set_outlet') as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            # Check if outlet is locked
            outlet_id = params['outlet'] - 1
            if self.outlet_locked[outlet_id]:
                return False, 'Outlet {} is locked. Cannot turn outlet on/off.'.format(params['outlet'])

            # Convert given state parameter to integer
            if params['state'] == 'on':
                state = 1
            else:
                state = 0

            # Issue SNMP SET command to given outlet
            outlet = [('IBOOTPDU-MIB', 'outletControl', outlet_id)]
            setcmd = yield self.snmp.set(outlet, self.version, state)
            self.log.info('{}'.format(setcmd))

        # Force SNMP GET status commands by rewinding the lastGet time by sample period
        self.lastGet = self.lastGet - self.sample_period

        return True, 'Set outlet {} to {}'.\
            format(params['outlet'], params['state'])

    @ocs_agent.param('outlet', choices=[1, 2, 3, 4, 5, 6, 7, 8])
    @ocs_agent.param('cycle_time', default=10, type=int)
    @inlineCallbacks
    def cycle_outlet(self, session, params=None):
        """cycle_outlet(outlet, cycle_time=10)

        **Task** - Cycle a particular outlet for given amount of seconds.

        Parameters
        ----------
        outlet : int
            Outlet number to cycle. Choices are 1-8 (physical outlets).
        cycle_time : int
            The amount of seconds to cycle an outlet. Default is 10 seconds.
        """
        with self.lock.acquire_timeout(3, job='cycle_outlet') as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            # Check if outlet is locked
            outlet_id = params['outlet'] - 1
            if self.outlet_locked[outlet_id]:
                return False, 'Outlet {} is locked. Cannot cycle outlet.'.format(params['outlet'])

            # Issue SNMP SET command for cycle time
            set_cycle = [('IBOOTPDU-MIB', 'outletCycleTime', outlet_id)]
            setcmd1 = yield self.snmp.set(set_cycle, self.version, params['cycle_time'])
            self.log.info('{}'.format(setcmd1))

            # Issue SNMP SET command to given outlet
            outlet = [('IBOOTPDU-MIB', 'outletControl', outlet_id)]
            setcmd2 = yield self.snmp.set(outlet, self.version, 2)
            self.log.info('{}'.format(setcmd2))
            self.log.info('Cycling outlet {} for {} seconds'.
                          format(outlet_id, params['cycle_time']))

        # Force SNMP GET status commands throughout the cycle time
        for i in range(params['cycle_time'] + 1):
            self.lastGet = self.lastGet - self.sample_period
            yield dsleep(1)

        return True, 'Cycled outlet {} for {} seconds'.\
            format(params['outlet'], params['cycle_time'])

    @ocs_agent.param('_')
    @inlineCallbacks
    def set_initial_state(self, session, params=None):
        """set_initial_state()

        **Task** - Set outlets to their initial states.

        Performs a software reboot. The outlets are then set to their
        respective initial states. This takes about 30 seconds.
        """
        with self.lock.acquire_timeout(3, job='reboot') as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            # Send SNMP SET command to set rebootSystem to True
            reboot = [('IBOOTPDU-MIB', 'rebootSystem', 0)]
            setcmd = yield self.snmp.set(reboot, self.version, 1)
            self.log.info('{}'.format(setcmd))

        # Force SNMP GET status commands
        self.lastGet = self.lastGet - self.sample_period

        return True, 'Rebooting. Outlets will be set to their initial states.'

    @ocs_agent.param('outlet', choices=[1, 2, 3, 4, 5, 6, 7, 8])
    @ocs_agent.param('lock', choices=[True, False])
    def lock_outlet(self, session, params=None):
        """lock_outlet(outlet, lock)

        **Task** - Lock/unlocks a particular outlet, preventing change of state.

        Parameters
        ----------
        outlet : int
            Outlet number to lock/unlock. Choices are 1-8 (physical outlets).
        lock : bool
            Set to true to lock, set to false to unlock
        """
        with self.lock.acquire_timeout(3, job='lock_outlet') as acquired:
            if not acquired:
                return False, "Could not acquire lock"
            # Lock/unlock specific outlet
            outlet = params['outlet'] - 1
            if params['lock']:
                self.outlet_locked[outlet] = True
            elif not params['lock']:
                self.outlet_locked[outlet] = False

        return True, 'Set outlet {} lock to {}'.\
            format(params['outlet'], params['lock'])


def add_agent_args(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--address", help="Address to listen to.")
    pgroup.add_argument("--port", default=161,
                        help="Port to listen on.")
    pgroup.add_argument("--snmp-version", default='2', choices=['1', '2', '3'],
                        help="SNMP version for communication. Must match "
                             + "configuration on the ibootbar.")
    pgroup.add_argument("--mode", default='acq', choices=['acq', 'test'])
    pgroup.add_argument("--lock-outlet", nargs='+', type=int,
                        help="List of outlets to lock on startup.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(agent_class='ibootbarAgent',
                                  parser=parser,
                                  args=args)

    if args.mode == 'acq':
        init_params = True
    elif args.mode == 'test':
        init_params = False

    agent, runner = ocs_agent.init_site_agent(args)
    p = ibootbarAgent(agent,
                      address=args.address,
                      port=int(args.port),
                      version=int(args.snmp_version),
                      lock_outlet=args.lock_outlet)

    agent.register_process("acq",
                           p.acq,
                           p._stop_acq,
                           startup=init_params, blocking=False)

    agent.register_task("set_outlet", p.set_outlet, blocking=False)
    agent.register_task("cycle_outlet", p.cycle_outlet, blocking=False)
    agent.register_task("set_initial_state", p.set_initial_state, blocking=False)
    agent.register_task("lock_outlet", p.lock_outlet, blocking=False)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



"""
temporary automation for telescope platforms.

"""

import argparse
import math
import random
import time

import ocs
import yaml
from ocs.ocs_client import OCSClient


def assert_ok(result):
    if result.status != ocs.base.OK:
        print('RESULT!=OK!', result)
        raise RuntimeError()


def wait_verbosely(target, timeout=5, msg=' ... still going '):
    last_stop = 0
    while True:
        # Watch for process exit
        result = target.wait(timeout=timeout)
        if result.status == ocs.base.TIMEOUT:
            print(msg, get_pos())
        elif result.status == ocs.base.ERROR:
            raise RuntimeError('Operation failed.')
        else:
            break
        # Detect az fault
        if get_faults()['az_summary'] and time.time() - last_stop > 10:
            print(' -- az summary fault detected, stop&clear.')
            assert_ok(c.stop_and_clear())
            last_stop = time.time()

    return True


def safe_get_status():
    # dodge a race condition
    for i in range(10):
        try:
            return c.monitor.status().session['data']['StatusDetailed']
        except KeyError:
            time.sleep(0.01)
    raise KeyError("Could not read StatusDetailed!")


def get_pos():
    status = safe_get_status()
    return (status['Azimuth current position'],
            status['Elevation current position'])


def get_faults():
    status = safe_get_status()
    return {
        'az_summary': status['Azimuth summary fault'],
        'local_mode': not status['ACU in remote mode'],
        'safe_lock': status['Safe'],
    }


def clear_faults():
    c.clear_faults.start()
    c.clear_faults.wait()


def set_boresight(angle):
    assert_ok(c.set_boresight.start(target=angle))
    wait_verbosely(c.set_boresight, msg=' ... setting boresight...')


def steps(targets=[], **kw):
    for az, el in targets:
        _az, _el = get_pos()
        if az is None:
            az = _az
        if el is None:
            el = _el
        assert_ok(c.go_to.start(az=az, el=el))
        wait_verbosely(c.go_to, msg=' ... moving')


def scan(az, el, throw,
         v_az=1.,
         a_az=1.,
         num_scans=3,
         step_time=1.,
         wait_to_start=10.,
         init='mid',
         init_az=None,
         ramp_up=0,
         ):
    if init_az is None:
        init_az = az
    print(f'Going to {init_az}, {el}')
    assert_ok(c.go_to.start(az=init_az, el=el))
    wait_verbosely(c.go_to, msg=' ... still go_to-ing')

    # assert(init == 'end')
    # ramp_up = az - throw - init_az
    # print(ramp_up)

    print('Checking positions ...')
    az1, el1 = get_pos()
    assert (abs(az1 - init_az) < .1 and abs(el1 - el) < .1)

    print('Stop and clear.')
    assert_ok(c.stop_and_clear())

    assert_ok(
        c.generate_scan.start(
            az_endpoint1=az - throw,
            az_endpoint2=az + throw,
            az_speed=v_az,
            az_accel=a_az,
            el_endpoint1=el,
            el_endpoint2=el,
            num_scans=num_scans,
            az_start=init)
    )

    wait_verbosely(c.generate_scan, msg=' ... still scanning ...')

    print('Finally we are at', get_pos())
    # time.sleep(2)
    time.sleep(5)

    print('Stop + clear')
    assert_ok(c.stop_and_clear())


def plan_scan(az, el, throw, v_az=1, a_az=1, init='end',
              full_ramp=False, num_scans=1):
    # Initialize arguments suitable for passing to scan() ...
    plan = {
        'az': az,
        'el': el,
        'throw': throw,
        'v_az': v_az,
        'a_az': a_az,
        'init': init,
        'init_az': az,
        'num_scans': num_scans,
    }
    info = {}

    # Point separation?  At least 5 points per leg, preferably 10.
    dt = 2 * abs(throw / v_az) / 10
    dt = min(max(dt, 0.1), 1.0)
    assert (2 * abs(throw / v_az) / dt >= 5)
    plan['step_time'] = dt

    # Turn around prep distance? 5 point periods, times the vel.
    az_prep = 5 * dt * v_az

    # Ramp-up distance needed
    a0 = 1.  # Peak accel of ramp-up...
    az_rampup = v_az**2 / a0
    info['az_prep'] = az_prep
    info['az_rampup'] = az_rampup

    # Any az ramp-up prep required?
    if full_ramp:
        ramp_up = az_rampup
    elif init == 'mid':
        ramp_up = max(az_prep + az_rampup - abs(throw), 0)
    elif init == 'end':
        ramp_up = max(az_prep + az_rampup - 2 * abs(throw), 0)
    else:
        raise  # init in ['mid', 'end']
    plan['ramp_up'] = ramp_up

    # Set wait time (this comes out a little lower than its supposed to...)
    # plan['wait_time'] = v_az / a0 * 2
    info['pre_time'] = v_az / a0
    plan['wait_to_start'] = max(5, info['pre_time'] * 1.2)

    # Fill out some other useful info...
    plan['init_az'] = az - math.copysign(ramp_up, throw)
    if init == 'end':
        plan['init_az'] -= throw

    info['total_time'] = (
        num_scans * (2 * abs(throw) / v_az + 2 * v_az / a_az)
        + ramp_up / v_az * 2
        + plan['wait_to_start'])

    return plan, info


class _Plan:
    code = 'something'

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        X = self.get_plan(self.index)
        self.index += 1
        return X


class ScanPlan(_Plan):
    pass


class PointPlan(_Plan):
    pass


class GridPlan(ScanPlan):
    DEFAULT_DURATION = 1200
    code = 'grid'

    def __init__(self, **kw):
        self.config = {
            'els': [50., 55., 60.],
            'azs': [0, 50, 90, 140, 180, 220, 270, 310],
            'az_throw': 20,
            'num_scans': 7,
            # 'num_scans': 1,
            'v_az': 1.,
            'a_az': 1.,
        }
        self.config.update(kw)
        self.index = 0

    def get_plan(self, index):
        c = self.config
        superphase = index // len(c['azs']) % 2
        step = 1 - 2 * superphase

        def get_mod(entry):
            return entry[index % len(entry)]

        plan, info = plan_scan(get_mod(c['azs'][::step]), get_mod(c['els']),
                               c['az_throw'] * step,
                               v_az=c['v_az'], a_az=c['a_az'],
                               num_scans=c['num_scans'])
        return plan, info


class ElNod(PointPlan):
    DEFAULT_DURATION = 60
    code = 'elnod'

    def __init__(self, **kw):
        self.repeat = kw.get('repeat', 1)

    def get_plan(self, index):
        t = {'targets': [(None, 20), (None, 90)] * self.repeat}
        return t, {}


class SchedulePlan(ScanPlan):
    DEFAULT_DURATION = 1200
    code = 'mock-sched'

    def __init__(self, sched_files, format='toast3',
                 dedup=True, dwell_time=None):
        if isinstance(sched_files, str):
            sched_files = [sched_files]
        if dwell_time is None:
            dwell_time = 300
        rows = []
        row_track = set()
        for filename in sched_files:
            skip = 3
            for line in open(filename):
                if skip > 0:
                    skip -= 1
                    continue
                w = line.split()
                row = {
                    'mjd0': float(w[4]),  # mjd start
                    'mjd1': float(w[5]),  # mjd end
                    'boresight': float(w[6]),  # boresight rotation
                    'az_min': float(w[8]),  # az min
                    'az_max': float(w[9]),  # az max
                    'el': float(w[10]),  # el
                }
                # rebranch
                row['az_max'] = row['az_min'] + (row['az_max'] - row['az_min']) % 360
                row['az'] = (row['az_max'] + row['az_min']) / 2
                row['az_throw'] = (row['az_max'] - row['az_min']) / 2

                # enrich
                row['v_az'] = 1. / math.cos(row['el'] * math.pi / 180)
                row['a_az'] = 2.
                row['num_scans'] = (dwell_time // (row['az_throw'] / row['v_az'])) + 1
                if dedup:
                    key = tuple([row[k] for k in ['az', 'az_throw', 'el']])
                    if key in row_track:
                        continue
                    row_track.add(key)
                rows.append(row)
        self.rows = rows

    def get_plan(self, index):
        row = self.rows[index % len(self.rows)]
        kw = {k: row[k] for k in ['v_az', 'a_az', 'num_scans']}
        return plan_scan(row['az'], row['el'], row['az_throw'],
                         **kw)


class GreasePlan(ScanPlan):
    DEFAULT_DURATION = 120
    code = 'grease'

    def __init__(self, phase=None):
        self.n = 4
        self.el_n = 9
        if phase is None:
            phase = int(random.random() * self.n)
        self.phase = phase

    def get_plan(self, index):
        phase = (self.phase + index) % self.n
        el = 30 + 40 * (index % self.el_n) / (self.el_n - 1)
        limits = [-80 + phase * 5, 440 + (phase - self.n - 1) * 5]
        print(limits)
        return plan_scan((limits[0] + limits[1]) / 2, el, (limits[1] - limits[0]) / 2,
                         v_az=2, a_az=0.5, num_scans=2, init='end')


def get_plan(config):
    if isinstance(config, str):
        config = yaml.safe_load(open(config, 'rb').read())
    for i, step in enumerate(config['steps']):
        if step['type'] == 'grease':
            driver = GreasePlan()
        elif step['type'] == 'schedule':
            driver = SchedulePlan(step['files'], dwell_time=step.get('dwell_time'))
        elif step['type'] == 'grid':
            driver = GridPlan()
        elif step['type'] == 'elnod':
            driver = ElNod(**step)
        else:
            raise ValueError('Invalid step type "%s"' % step['type'])
        plan = {'driver': driver,
                'duration': step.get('duration', driver.DEFAULT_DURATION),
                }
        config['steps'][i] = plan  # replace
    return config


def set_client(instance_id):
    global c
    c = OCSClient(instance_id)
    return c


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config')
    parser.add_argument('instance_id', help="e.g. acu-sat1")
    parser.add_argument('--hours', type=float)
    args = parser.parse_args()

    if args.config is not None:
        # Get the plan.
        super_plan = get_plan(args.config)

    c = OCSClient(args.instance_id)
    now = time.time()
    stop_at = None
    if args.hours:
        stop_at = now + args.hours * 3600.

    plan_idx = 1
    while stop_at is None or time.time() < stop_at:
        t0 = time.time()
        t = t0
        active_plan = super_plan['steps'][plan_idx]
        print(active_plan['driver'])
        for plan, info in active_plan['driver']:
            if isinstance(plan, ScanPlan):
                scan(**plan)
            elif isinstance(plan, PointPlan):
                steps(**plan)
            else:
                raise ValueError
            if time.time() - t0 > active_plan['duration']:
                break
        plan_idx += 1
        if plan_idx >= len(super_plan['steps']):
            if super_plan.get('loop', True):
                plan_idx = 0



import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from twisted.internet import reactor

import socs.agents.hwp_pmx.drivers.PMX_ethernet as pmx

txaio.use_twisted()


class HWPPMXAgent:
    """Agent for interfacing with a PMX Kikusui power supply
    to control the current and voltage that drive the rotation of the CHWP.

    Args:
        ip (str): IP address for the PMX Kikusui power supply
        port (str): Port for the PMX Kikusui power supply
        f_sample (float): sampling frequency (Hz)
    """

    def __init__(self, agent, ip, port, f_sample=1):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.ip = ip
        self.port = port

        self.f_sample = f_sample

        self._initialized = False
        self.take_data = False

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'hwppmx', record=True, agg_params=agg_params, buffer_time=1)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init_connection(self, session, params=None):
        """init_connection(auto_acquire=False)
        **Task** - Initialize connection to PMX

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.
        """
        if self._initialized:
            self.log.info("Connection already initialized. Returning...")
            return True, "Connection already initialized"

        with self.lock.acquire_timeout(0, job='init_connection') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run init_connection because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            try:
                self.dev = pmx.PMX(ip=self.ip, port=self.port)
                self.log.info('Connected to PMX Kikusui')
            except BrokenPipeError:
                self.log.error('Could not establish connection to PMX Kikusui')
                reactor.callFromThread(reactor.stop)
                return False, 'Unable to connect to PMX Kikusui'

        if params['auto_acquire']:
            self.agent.start('acq')

        return True, 'Connection to PMX established'

    def set_on(self, session, params):
        """set_on()
        **Task** - Turn on the PMX Kikusui.
        """
        with self.lock.acquire_timeout(3, job='set_on') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set on because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            self.dev.turn_on()
        return True, 'Set PMX Kikusui on'

    def set_off(self, session, params):
        """set_off()
        **Task** - Turn off the PMX Kikusui.
        """
        with self.lock.acquire_timeout(3, job='set_off') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set on because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            self.dev.turn_off()
        return True, 'Set PMX Kikusui off'

    @ocs_agent.param('curr', default=0, type=float, check=lambda x: 0 <= x <= 3)
    def set_i(self, session, params):
        """set_i(curr=0)
        **Task** - Set the current.

        Parameters:
            curr (float): set current
        """
        with self.lock.acquire_timeout(3, job='set_i') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set i because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            msg, val = self.dev.set_current(params['curr'])
        return True, msg

    @ocs_agent.param('volt', default=0, type=float, check=lambda x: 0 <= x <= 35)
    def set_v(self, session, params):
        """set_v(volt=0)
        **Task** - Set the voltage.

        Parameters:
            volt (float): set voltage
        """
        with self.lock.acquire_timeout(3, job='set_v') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set v because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            msg, val = self.dev.set_voltage(params['volt'])
        return True, msg

    @ocs_agent.param('curr', default=1., type=float, check=lambda x: 0. <= x <= 3.)
    def set_i_lim(self, session, params):
        """set_i_lim(curr=1)
        **Task** - Set the drive current limit.

        Parameters:
            curr (float): limit current
        """
        with self.lock.acquire_timeout(3, job='set_i_lim') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set v lim because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            msg = self.dev.set_current_limit(params['curr'])
        return True, msg

    @ocs_agent.param('volt', default=32., type=float, check=lambda x: 0. <= x <= 35.)
    def set_v_lim(self, session, params):
        """set_v_lim(volt=32)
        **Task** - Set the drive voltage limit.

        Parameters:
            volt (float): limit voltage
        """
        with self.lock.acquire_timeout(3, job='set_v_lim') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set v lim because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            msg = self.dev.set_voltage_limit(params['volt'])
        return True, msg

    def use_ext(self, session, params):
        """use_ext()
        **Task** - Set the PMX Kikusui to use an external voltage control. Doing so
        enables PID control.
        """
        with self.lock.acquire_timeout(3, job='use_ext') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not use external voltage because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            self.dev.use_external_voltage()
        return True, 'Set PMX Kikusui to PID control'

    def ign_ext(self, session, params):
        """ign_ext()
        **Task** - Set the PMX Kiksui to ignore external voltage control. Doing so
        disables the PID and switches to direct control.
        """
        with self.lock.acquire_timeout(3, job='ign_ext') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not ignore external voltage because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            self.dev.ign_external_voltage()
        return True, 'Set PMX Kikusui to direct control'

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params):
        """acq(test_mode=False)

        **Process** - Start data acquisition.

        Parameters:
            test_mode (bool, optional): Run the Process loop only once.
                This is meant only for testing. Default is False.

        Notes:
            The most recent data collected is stored in session data in the
            structure::

                >>> response.session['data']
                {'curr': 0,
                 'volt': 0,
                 'last_updated': 1649085992.719602}

        """
        sleep_time = 1 / self.f_sample - 0.01

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running"
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')
            self.take_data = True

            last_release = time.time()
            while self.take_data:
                current_time = time.time()
                if current_time - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire lock, currently held by {self.lock.job}.")
                        continue

                data = {
                    'timestamp': current_time,
                    'block_name': 'hwppmx',
                    'data': {}
                }
                msg, curr = self.dev.meas_current()
                data['data']['current'] = curr
                msg, volt = self.dev.meas_voltage()
                data['data']['voltage'] = volt

                self.agent.publish_to_feed('hwppmx', data)
                session.data = {'curr': curr,
                                'volt': volt,
                                'last_updated': current_time}

                time.sleep(sleep_time)

                if params['test_mode']:
                    break

            self.agent.feeds['hwppmx'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stop acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip', type=str,
                        help="ip address for kikusui PMX")
    pgroup.add_argument('--port', type=int,
                        help="port for kikusui PMX")
    pgroup.add_argument('--mode', type=str, default='acq', choices=['init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument('--sampling-frequency', type=float,
                        help="Sampling frequency for data acquisition")
    return parser


def main(args=None):
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPPMXAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    kwargs = {'ip': args.ip, 'port': args.port}
    if args.sampling_frequency is not None:
        kwargs['f_sample'] = args.sampling_frequency
    PMX = HWPPMXAgent(agent, **kwargs)

    agent.register_task('init_connection', PMX.init_connection, startup=init_params)
    agent.register_process('acq', PMX.acq, PMX._stop_acq)
    agent.register_task('set_on', PMX.set_on)
    agent.register_task('set_off', PMX.set_off)
    agent.register_task('set_i', PMX.set_i)
    agent.register_task('set_v', PMX.set_v)
    agent.register_task('set_i_lim', PMX.set_i_lim)
    agent.register_task('set_v_lim', PMX.set_v_lim)
    agent.register_task('use_ext', PMX.use_ext)
    agent.register_task('ign_ext', PMX.ign_ext)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import calendar
import datetime
import math
import time

import numpy as np

#: The number of seconds in a day.
DAY = 86400


def constant_velocity_scanpoints(azpts, el, azvel, acc, ntimes):
    """
    Produces lists of times, azimuths, elevations, azimuthal velocities,
    elevation velocities, azimuth motion flags, and elevation motion flags
    for a finitely long azimuth scan with constant velocity. Scan begins
    at the first azpts value.

    Parameters:
        azpts (2-tuple): The endpoints of motion in azimuth, where the
            first point is the start position of the scan.
        el (float): The elevation that is maintained throughout the scan
        azvel (float): Desired speed of the azimuth motion in degrees/sec
        acc (float): The turnaround acceleration in degrees/sec^2
        ntimes(int): Number of times to travel between the endpoints.
            ntimes = 1 corresponds to a scan from, ex., left to right, and does not
            return to left.

    Returns:
        tuple of lists : (times, azimuths, elevations, azimuth veolicities,
        elevation velocities, azimuth flags, elevation flags)
    """
    if float(azvel) == 0.0:
        print('Azimuth velocity is zero, invalid scan parameter')
        return False
    if float(acc) == 0.0:
        print('Acceleration is zero, unable to calculate turnaround')
        return False
    turn_time = 2 * azvel / acc
    tot_time_dir = float((abs(azpts[1] - azpts[0])) / azvel)
    num_dirpoints = int(tot_time_dir * 10.)
    if num_dirpoints < 2:
        print('Scan is too short to run')
        return False
    sect_start_time = 0.0
    conctimes = []
    concaz = []
    el1 = np.linspace(el, el, num_dirpoints)
    concel = list(el1)
    concva = []
    ve1 = np.zeros(num_dirpoints)
    concve = list(ve1)

    # Flag values:
    # 0 : unidentified portion of the scan
    # 1 : constant velocity, with the next point at the same velocity
    # 2 : final point before a turnaround
    azflags = [1 for i in range(num_dirpoints - 1)]
    azflags += [2]
    all_azflags = []

    elflags = [0 for i in range(num_dirpoints)]
    all_elflags = []

    for n in range(ntimes):
        # print(str(n)+' '+str(sect_start_time))
        end_dir_time = sect_start_time + tot_time_dir
        time_for_section = np.linspace(sect_start_time, end_dir_time,
                                       num_dirpoints)
        if n % 2 != 0:
            new_az = np.linspace(azpts[1], azpts[0], num_dirpoints)
            new_va = np.zeros(num_dirpoints) + \
                np.sign(azpts[0] - azpts[1]) * azvel
        else:
            new_az = np.linspace(azpts[0], azpts[1], num_dirpoints)
            new_va = np.zeros(num_dirpoints) + \
                np.sign(azpts[1] - azpts[0]) * azvel

        conctimes.extend(time_for_section)
        concaz.extend(new_az)
        concel.extend(el1)
        concva.extend(new_va)
        concve.extend(ve1)
        all_azflags.extend(azflags)
        all_elflags.extend(elflags)
        sect_start_time = time_for_section[-1] + turn_time

    all_azflags[-1] = 0

    return conctimes, concaz, concel, concva, concve, all_azflags, \
        all_elflags


def from_file(filename):
    """
    Produces properly formatted lists of times, azimuth and elevation
    locations, azimuth and elevation velocities, and azimuth and elevation
    motion flags for a finitely long scan from a numpy file. Numpy file
    must be formatted as an array of arrays in the order [times, azimuths,
    elevations, azimuth velocities, elevation velocities]

    Parameters:
        filename (str): Full path to the numpy file containing scan
            parameter array

    Returns:
        tuple of lists: (times, azimuths, elevations, azimuth velocities,
        elevation velocities, azimuth flags, elevation flags)

    NOTE: Flags can be set in the numpy file (0=unspecified, 1=constant
    velocity, 2=last point before turnaround). If flags are not set in
    the file, all flags are set to 0 to accommodate non-linear scans
    """
    info = np.load(filename)
    if len(info) not in [5, 7]:
        raise ValueError(f'Unexpected field count ({len(info)}) in {filename}')
    conctimes = info[0]
    concaz = info[1]
    concel = info[2]
    concva = info[3]
    concve = info[4]
    if len(info) == 5:
        az_flags = np.zeros(len(conctimes), int)
        el_flags = az_flags
    elif len(info) == 7:
        az_flags = info[5].astype('int')
        el_flags = info[6].astype('int')
    return conctimes, concaz, concel, concva, concve, az_flags, el_flags


def ptstack_format(conctimes, concaz, concel, concva, concve, az_flags,
                   el_flags, start_offset=0, absolute=False):
    """
    Produces a list of lines in the format necessary to upload to the ACU
    to complete a scan. Params are the outputs of from_file,
    constant_velocity_scanpoints, or generate_constant_velocity_scan.

    Parameters:
        conctimes (list): Times starting at 0 for the ACU to reach
            associated positions
        concaz (list): Azimuth positions associated with conctimes
        concel (list): Elevation positions associated with conctimes
        concva (list): Azimuth velocities associated with conctimes
        concve (list): Elevation velocities associated with conctimes
        az_flags (list): Flags associated with azimuth motions at
            conctimes
        el_flags (list): Flags associated with elevation motions at
            conctimes
        start_offset (float): Offset, in seconds, to apply to all
            timestamps.
        absolute (bool): If true, timestamps are taken at face value,
            and only start_offset is added.  If false, then the current
            time is also added (but note that if the first timestamp
            is 0, then you will need to also pass start_offset > 0).

    Returns:
        list: Lines in the correct format to upload to the ACU
    """

    fmt = '%j, %H:%M:%S'
    if not absolute:
        start_offset = time.time() + start_offset
    true_times = [start_offset + i for i in conctimes]
    fmt_times = [time.strftime(fmt, time.gmtime(t))
                 + ('{tt:.6f}'.format(tt=t % 1.))[1:] for t in true_times]

    all_lines = [('{ftime}; {az:.6f}; {el:.6f}; {azvel:.4f}; '
                  '{elvel:.4f}; {azflag}; {elflag}'
                  '\r\n'.format(ftime=fmt_times[n], az=concaz[n],
                                el=concel[n], azvel=concva[n], elvel=concve[n],
                                azflag=az_flags[n], elflag=el_flags[n]))
                 for n in range(len(fmt_times))]

    return all_lines


def timecode(acutime, now=None):
    """Takes the time code produced by the ACU status stream and returns
    a unix timestamp.

    Parameters:
        acutime (float): The time recorded by the ACU status stream,
            corresponding to the fractional day of the year.
        now (float): The time, as unix timestamp, to assume it is now.
            This is for testing, it defaults to time.time().

    """
    sec_of_day = (acutime - 1) * DAY
    if now is None:
        now = time.time()  # testing

    # This guard protects us at end of year, when time.time() and
    # acutime might correspond to different years.
    if acutime > 180:
        context = datetime.datetime.utcfromtimestamp(now - 30 * DAY)
    else:
        context = datetime.datetime.utcfromtimestamp(now + 30 * DAY)

    year = context.year
    gyear = calendar.timegm(time.strptime(str(year), '%Y'))
    comptime = gyear + sec_of_day
    return comptime


def generate_constant_velocity_scan(az_endpoint1, az_endpoint2, az_speed,
                                    acc, el_endpoint1, el_endpoint2,
                                    el_speed=0,
                                    num_batches=None,
                                    num_scans=None,
                                    start_time=None,
                                    wait_to_start=10.,
                                    step_time=1.,
                                    batch_size=500,
                                    az_start='mid_inc',
                                    az_first_pos=None,
                                    ptstack_fmt=True):
    """Python generator to produce times, azimuth and elevation positions,
    azimuth and elevation velocities, azimuth and elevation flags for
    arbitrarily long constant-velocity azimuth scans.

    Parameters:
        az_endpoint1 (float): azimuth endpoint for the scan start
        az_endpoint2 (float): second azimuth endpoint of the scan
        az_speed (float): speed of the constant-velocity azimuth motion
        acc (float): turnaround acceleration for the azimuth motion at the
            endpoints
        el_endpoint1 (float): elevation endpoint for the scan start
        el_endpoint2 (float): second elevation endpoint of the scan. For
            constant az scans, this must be equal to el_endpoint1.
        el_speed (float): speed of the elevation motion. For constant az
            scans, set to 0.0
        num_batches (int or None): sets the number of batches for the
            generator to create. Default value is None (interpreted as infinite
            batches).
        num_scans (int or None): if not None, limits the points
          returned to the specified number of constant velocity legs.
        start_time (float or None): a ctime at which to start the scan.
            Default is None, which is interpreted as starting now +
            wait_to_start.
        wait_to_start (float): number of seconds to wait between
            start_time and when the scan actually starts. Default is 10 seconds.
        step_time (float): time between points on the constant-velocity
            parts of the motion. Default value is 1.0 seconds. Minimum value is
            0.05 seconds.
        batch_size (int): number of values to produce in each iteration.
            Default is 500. Batch size is reset to the length of one leg of the
            motion if num_batches is not None.
        az_start (str): part of the scan to start at.  To start at one
            of the extremes, use 'az_endpoint1', 'az_endpoint2', or
            'end' (same as 'az_endpoint1').  To start in the midpoint
            of the scan use 'mid_inc' (for first half-leg to have
            positive az velocity), 'mid_dec' (negative az velocity),
            or 'mid' (velocity oriented towards endpoint2).
        az_first_pos (float): If not None, the first az scan will
            start at this position (but otherwise proceed in the same
            starting direction).
        ptstack_fmt (bool): determine whether values are produced with the
            necessary format to upload to the ACU. If False, this function will
            produce lists of time, azimuth, elevation, azimuth velocity,
            elevation velocity, azimuth flags, and elevation flags. Default is
            True.

    """
    az_min = min(az_endpoint1, az_endpoint2)
    az_max = max(az_endpoint1, az_endpoint2)
    if az_max == az_min:
        raise ValueError('Generator requires two different az endpoints!')
    if az_start in ['az_endpoint1', 'az_endpoint2', 'end']:
        if az_start in ['az_endpoint1', 'end']:
            az = az_endpoint1
        else:
            az = az_endpoint2
        increasing = (az == az_min)
    elif az_start in ['mid_inc', 'mid_dec', 'mid']:
        az = (az_endpoint1 + az_endpoint2) / 2
        if az_start == 'mid':
            increasing = az_endpoint2 > az_endpoint1
        elif az_start == 'mid_inc':
            increasing = True
        else:
            increasing = False
    else:
        raise ValueError('az_start value not supported. Choose from '
                         'az_endpoint1, az_endpoint2, mid_inc, mid_dec')
    az_vel = az_speed if increasing else -az_speed

    # Bias the starting point for the first leg?
    if az_first_pos is not None:
        az = az_first_pos

    if start_time is None:
        t0 = time.time() + wait_to_start
    else:
        t0 = start_time
    t = 0
    turntime = 2.0 * az_speed / acc
    el = el_endpoint1
    if step_time < 0.05:
        raise ValueError('Time step size too small, must be at least '
                         '0.05 seconds')
    daz = step_time * az_speed
    el_vel = el_speed
    az_flag = 0
    el_flag = 0
    if num_batches is None:
        stop_iter = float('inf')
    else:
        stop_iter = num_batches
        batch_size = int(np.ceil((az_max - az_min) / daz))

    def dec_num_scans():
        nonlocal num_scans
        if num_scans is not None:
            num_scans -= 1

    def check_num_scans():
        return num_scans is None or num_scans > 0

    i = 0
    while i < stop_iter and check_num_scans():
        i += 1
        point_block = [[], [], [], [], [], [], []]
        for j in range(batch_size):
            point_block[0].append(t + t0)
            point_block[1].append(az)
            point_block[2].append(el)
            point_block[3].append(az_vel)
            point_block[4].append(el_vel)
            point_block[5].append(az_flag)
            point_block[6].append(el_flag)
            t += step_time

            if increasing:
                if az <= (az_max - 2 * daz):
                    az += daz
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = True
                elif az == az_max:
                    t += turntime
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = False
                    dec_num_scans()
                else:
                    az_remaining = az_max - az
                    time_remaining = az_remaining / az_speed
                    az = az_max
                    t += (time_remaining - step_time)
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0
                    increasing = True
            else:
                if az >= (az_min + 2 * daz):
                    az -= daz
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = False
                elif az == az_min:
                    t += turntime
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = True
                    dec_num_scans()
                else:
                    az_remaining = az - az_min
                    time_remaining = az_remaining / az_speed
                    az = az_min
                    t += (time_remaining - step_time)
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0
                    increasing = False

            if not check_num_scans():
                # Kill the velocity on the last point and exit -- this
                # was recommended at LAT FAT for smoothly stopping the
                # motino at end of program.
                point_block[3][-1] = 0
                point_block[4][-1] = 0
                break

        if ptstack_fmt:
            yield ptstack_format(point_block[0], point_block[1],
                                 point_block[2], point_block[3],
                                 point_block[4], point_block[5],
                                 point_block[6],
                                 start_offset=3, absolute=True)
        else:
            yield (point_block[0], point_block[1], point_block[2],
                   point_block[3], point_block[4], point_block[5],
                   point_block[6])


def plan_scan(az_end1, az_end2, el, v_az=1, a_az=1, az_start=None):
    """Determine some important parameters for running a ProgramTrack
    scan with the desired end points, velocity, and mean turn-around
    acceleration.

    These get complicated in the limit of high velocity and narrow
    scan.

    Returns:

      A dict with outputs of the calculations.  The following items
      must be considered when generating and posting the track points:

      - 'step_time': The recommended track point separation, in
        seconds.
      - 'wait_to_start': The minimum time (s) between initiating
        ProgramTrack mode and the first uploaded point's timestamp.
      - 'init_az': The az (deg) at which to position the telescope
        before beginning the scan.  This takes into account any "ramp
        up" that needs to occur and the fact that such ramp up needs
        to be finished before the ACU starts profiling the first
        turn-around.

      The following dict items provide additional detail /
      intermediate results:

      - 'scan_start_buffer': Minimum amount (deg of az) by which to
        shift the start of the first scan leg in order to satisfy the
        requirements for az_prep and az_rampup.  This ultimately is
        what can make init_az different from the natural first leg
        starting point.  This parameter is always non-negative.
      - 'turnprep_buffer': Minimum azimuth travel required for
        ProgramTrack to prepare a turn-around.
      - 'rampup_buffer': Minimum azimuth travel required for
        ProgramTrack to ramp up to the first leg velocity.  Degrees,
        positive.
      - 'rampup_time': Number of seconds before the first track point
        where the platform could start moving (as part of smooth
        acceleration into the initial velocity).

    """
    # Convert Agent-friendly arguments to az/throw/init
    az = (az_end1 + az_end2) / 2
    throw = (az_end2 - az_end1) / 2

    if az_start in [None, 'mid']:
        init = 'mid'
    elif az_start == 'mid_inc':
        init = 'mid'
        throw = abs(throw)
    elif az_start == 'mid_dec':
        init = 'mid'
        throw = -abs(throw)
    elif az_start in ['az_endpoint1', 'end']:
        init = 'end'
    elif az_start in ['az_endpoint2']:
        init = 'end'
        throw = -throw
    else:
        raise ValueError(f'Unexpected az_start={az_start}')

    # Info to pass back.
    plan = {}

    # Point time separation: at least 5 points per leg, preferably 10.
    dt = 2 * abs(throw / v_az) / 10
    dt = min(max(dt, 0.1), 1.0)
    assert (2 * abs(throw / v_az) / dt >= 5)
    plan['step_time'] = dt

    # Turn around prep distance (deg)? 5 point periods, times the vel.
    turnprep_buffer = 5 * dt * v_az

    # Ramp-up distance needed
    a0 = 1.  # Peak accel of ramp-up...
    rampup_buffer = v_az**2 / a0
    plan['turnprep_buffer'] = turnprep_buffer
    plan['rampup_buffer'] = rampup_buffer

    # Any az ramp-up prep required?
    if init == 'mid':
        scan_start_buffer = max(turnprep_buffer + rampup_buffer - abs(throw), 0)
    elif init == 'end':
        scan_start_buffer = max(turnprep_buffer + rampup_buffer - 2 * abs(throw), 0)
    plan['scan_start_buffer'] = scan_start_buffer

    # Set wait time (this comes out a little lower than its supposed to...)
    # plan['wait_time'] = v_az / a0 * 2
    plan['rampup_time'] = v_az / a0
    plan['wait_to_start'] = max(5, plan['rampup_time'] * 1.2)

    # Fill out some other useful info...
    plan['init_az'] = az - math.copysign(scan_start_buffer, throw)
    if init == 'end':
        plan['init_az'] -= throw

    return plan



import time
from socket import AF_INET, SOCK_STREAM, socket


class PMX:
    """The PMX object for communicating with the Kikusui PMX power supplies.
    Args:
        tcp_ip (str): TCP IP address
        tcp_port (int): TCP port
    """

    def __init__(self, ip, port):
        self.sock = socket(AF_INET, SOCK_STREAM)
        self.sock.connect((ip, port))
        self.sock.settimeout(5)

        self.wait_time = 0.01
        self.buffer_size = 128

    def close(self):
        self.sock.close()

    def read(self):
        return self.sock.recv(self.buffer_size).decode('utf-8')

    def wait(self):
        time.sleep(self.wait_time)

    def check_output(self):
        """ Return the output status """
        self.sock.sendall(b'output?\n')
        val = int(self.read())
        msg = "Measured output state = "
        states = {0: 'OFF', 1: 'ON'}
        if val in states:
            msg += states[val]
        else:
            msg += 'Fail'
        return msg, val

    def turn_on(self):
        """ Turn the PMX on """
        self.sock.sendall(b'output 1\n')
        self.wait()
        return self.check_output()

    def turn_off(self):
        """ Turn the PMX off """
        self.sock.sendall(b'output 0\n')
        self.wait()
        return self.check_output()

    def check_current(self):
        """ Check the current setting """
        self.sock.sendall(b'curr?\n')
        val = float(self.read())
        msg = "Current setting = {:.3f} A".format(val)
        return msg, val

    def check_voltage(self):
        """ Check the voltage setting """
        self.sock.sendall(b'volt?\n')
        val = float(self.read())
        msg = "Voltage setting = {:.3f} V".format(val)
        return msg, val

    def meas_current(self):
        """ Measure the current """
        self.sock.sendall(b'meas:curr?\n')
        val = float(self.read())
        msg = "Measured current = {:.3f} A".format(val)
        return msg, val

    def meas_voltage(self):
        """ Measure the voltage """
        self.sock.sendall(b'meas:volt?\n')
        val = float(self.read())
        msg = "Measured voltage = {:.3f} V".format(val)
        return msg, val

    def set_current(self, curr):
        """ Set the current """
        self.sock.sendall(b'curr %a\n' % curr)
        self.wait()
        return self.check_current()

    def set_voltage(self, vol):
        """ Set the voltage """
        self.sock.sendall(b'volt %a\n' % vol)
        self.wait()
        return self.check_voltage()

    def check_source(self):
        """ Check the source of PMX """
        self.sock.sendall(b'volt:ext:sour?\n')
        val = self.read()
        msg = "Source: " + val
        return msg

    def use_external_voltage(self):
        """ Set PMX to use external voltage """
        self.sock.sendall(b'volt:ext:sour volt\n')
        self.wait()
        return self.check_source()

    def ign_external_voltage(self):
        """ Set PMX to ignore external voltage """
        self.sock.sendall(b'volt:ext:sour none\n')
        self.wait()
        return self.check_source()

    def set_current_limit(self, curr_lim):
        """ Set the PMX current limit """
        self.sock.sendall(b'curr:prot %a\n' % curr_lim)
        self.wait()
        self.sock.sendall(b'curr:prot?\n')
        val = float(self.read())
        msg = "Current Limit: {:.3f} A".format(val)
        return msg

    def set_voltage_limit(self, vol_lim):
        """ Set the PMX voltage limit """
        self.sock.sendall(b'volt:prot %a\n' % vol_lim)
        self.wait()
        self.sock.sendall(b'volt:prot?\n')
        val = float(self.read())
        msg = "Voltage Limit: {:.3f} V".format(val)
        return msg



import argparse
import os
import time

import txaio
from epics import PV
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker

diagnostic_root = 'TPG:SMRF:1'
regs = [
    'COUNTPLL', 'COUNT186M', 'COUNTSYNCERR', 'COUNTINTV', 'COUNTBRT',
    'COUNTTXCLK', 'DELTATXCLK', 'RATETXCLK'
]
regs_full = {
    r: f'{diagnostic_root}:{r}' for r in regs
}


class SmurfTimingCardAgent:
    """
    Agent to monitor diagnostic variables for the SMuRF Timing Card.
    """

    def __init__(self, agent, args):
        self.agent = agent
        self.log = agent.log
        self.sleep_time = args.sleep_time
        self.pvs = {
            k: PV(r) for k, r in regs_full.items()
        }
        self.timeout = args.timeout
        self.use_monitor = args.use_monitor
        agent.register_feed('diagnostics', record=True)

    def run(self, session, params):
        """run()

        **Process** -- Main loop for the timing card. This queries diagnostic
        PVs and publishes data.

        Notes:
            An example of the session data::

                >>> response.session['data']
                   {'COUNT186M': 1671196546,
                    'COUNTBRT': 480000,
                    'COUNTINTV': 122879999,
                    'COUNTPLL': 1,
                    'COUNTSYNCERR': 4,
                    'COUNTTXCLK': 1461994265,
                    'DELTATXCLK': 15416942,
                    'RATETXCLK': 123.33553599999999,
                    'timestamp': 1678983237.832111}

        """
        pacemaker = Pacemaker(1. / self.sleep_time)
        session.set_status('running')
        while session.status in ['starting', 'running']:
            data = {}
            for name, pv in self.pvs.items():
                res = pv.get(timeout=self.timeout, use_monitor=self.use_monitor)
                if res is None:
                    self.log.error("Timeout for PV: {name}", name=name)
                    break
                data[name] = res
            else:
                session.data = data
                session.data['timestamp'] = time.time()
                msg = {
                    'timestamp': time.time(),
                    'block_name': 'diagnostics',
                    'data': data
                }
                self.agent.publish_to_feed('diagnostics', msg)
            pacemaker.sleep()

        return True, 'Run process has been stopped'

    def _stop(self, session, params=None):
        session.set_status('stopping')


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--timeout', type=float, default=3.,
                        help='Timeout for epics caget calls')
    pgroup.add_argument('--sleep-time', type=float, default=30.,
                        help="Time to sleep between loop iterations")
    pgroup.add_argument('--use-monitor', action='store_true',
                        help="Use the epics monitored value. If False, will "
                             "re-poll for every get call.")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args("SmurfTimingCardAgent", parser=parser, args=args)
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))
    agent, runner = ocs_agent.init_site_agent(args)
    timing = SmurfTimingCardAgent(agent, args)
    agent.register_process('run', timing.run, timing._stop, startup=True)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()
